"""Models the MINT DICOM exchange protocol
See:
https://code.google.com/archive/p/medical-imaging-network-transport/downloads
"""
import datetime
from typing import ClassVar, List, Optional, Sequence, Set
from xml.etree import ElementTree
from xml.etree.ElementTree import ParseError

from pydantic.class_validators import validator
from pydantic.main import BaseModel
from pydicom.dataelem import DataElement
from pydicom.dataset import Dataset

from dicomtrolley.core import DICOMObject, Instance, Searcher, Series, Study
from dicomtrolley.exceptions import DICOMTrolleyException
from dicomtrolley.fields import (
    InstanceLevel,
    SeriesLevel,
    SeriesLevelPromotable,
    StudyLevel,
)


class QueryLevels:
    STUDY = "STUDY"
    SERIES = "SERIES"
    INSTANCE = "INSTANCE"

    ALL = {STUDY, SERIES, INSTANCE}


class MintObject(DICOMObject):
    """Python representation of MINT xml"""

    xml_element: ClassVar[str]

    def all_instances(self):
        """
        Returns
        -------
        List[MintInstance]
            All instances contained in this object
        """
        raise NotImplementedError()


class MintInstance(Instance, MintObject):
    xml_element: ClassVar = "{http://medical.nema.org/mint}instance"
    parent: "MintSeries"

    @classmethod
    def init_from_element(cls, element, parent):
        data = parse_attribs(element)

        return cls(data=data, uid=data.SOPInstanceUID, parent=parent)


class MintSeries(Series, MintObject):
    xml_element: ClassVar = "{http://medical.nema.org/mint}series"
    instances: List[MintInstance]
    parent: "MintStudy"

    @classmethod
    def init_from_element(cls, element, parent):
        data = parse_attribs(element)

        series = cls(
            data=data,
            uid=data.SeriesInstanceUID,
            parent=parent,
            instances=[],
        )
        for x in element.findall(MintInstance.xml_element):
            series.instances.append(
                MintInstance.init_from_element(x, parent=series)
            )

        return series


class MintStudy(Study, MintObject):
    xml_element: ClassVar = "{http://medical.nema.org/mint}study"

    mint_uuid: str  # non-DICOM MINT-only id for this series
    last_modified: str
    series: List[MintSeries]

    @classmethod
    def init_from_element(cls, element):
        data = parse_attribs(element)

        study = cls(
            data=data,
            uid=data.StudyInstanceUID,
            mint_uuid=element.attrib["studyUUID"],
            last_modified=element.attrib["lastModified"],
            series=[],
        )
        for x in element.findall(MintSeries.xml_element):
            study.series.append(MintSeries.init_from_element(x, parent=study))

        return study

    def dump_content(self) -> str:
        """Dump entire contents of this study and containing series, instances

        For quick inspection in scripts
        """

        output = [str(self.data)]
        for series in self.series:
            output.append(str(series.data))
            for instance in series.instances:
                output.append(str(instance.data))
        return "\n".join(output)


class MintAttribute:
    """A DICOM element like PatientID=001"""

    xml_element = "{http://medical.nema.org/mint}attr"


class MintQuery(BaseModel):
    """Things you can search for with the MINT find DICOM studies function

    Notes
    -----
    * All string arguments support (*) as a wildcard
    * non-pep8 parameter naming format follows MINT url parameters naming
    """

    # String parameters (wildcards allowed)
    studyInstanceUID: str = ""
    accessionNumber: str = ""
    patientName: str = ""
    patientID: str = ""
    modalitiesInStudy: str = ""
    institutionName: str = ""
    patientSex: str = ""
    studyDescription: str = ""
    institutionalDepartmentName: str = ""

    # date search parameters
    patientBirthDate: Optional[datetime.date]
    minStudyDate: Optional[datetime.datetime]
    maxStudyDate: Optional[datetime.datetime]

    # meta parameters: how to return results
    queryLevel: str = QueryLevels.STUDY  # to which depth to return results
    includeFields: List[str] = []  # which dicom fields to return
    limit: int = 0  # how many results to return. 0 = all

    class Config:
        extra = "forbid"  # raise ValueError when passing an unknown keyword to init

    def __str__(self):
        return str(self.as_parameters())

    @validator("maxStudyDate", always=True)
    def min_max_study_date_xor(cls, value, values):  # noqa: B902, N805
        """Min and max should both be given or both be empty"""
        if values.get("minStudyDate", None) and not value:
            raise ValueError(
                f"minStudyDate parameter was passed "
                f"({values['minStudyDate']}), "
                f"but maxStudyDate was not. Both need to be given"
            )
        elif value and not values.get("minStudyDate", None):
            raise ValueError(
                f"maxStudyDate parameter was passed ({value}), "
                f"but minStudyDate was not. Both need to be given"
            )
        return value

    @validator("includeFields")
    def include_fields_check(cls, include_fields, values):  # noqa: B902, N805
        """Include fields should be valid and match query level"""
        query_level = values["queryLevel"]
        valid_fields = get_valid_fields(query_level=query_level)
        for field in include_fields:
            if field not in valid_fields:
                raise ValueError(
                    f'"{field}" is not a valid include field for query '
                    f"level {query_level}. Valid fields: {valid_fields}"
                )
        return include_fields

    def as_parameters(self):
        """All non-empty query parameters. For use as url parameters"""
        parameters = {x: y for x, y in self.dict().items() if y}

        if "minStudyDate" in parameters:
            parameters["minStudyDate"] = parameters["minStudyDate"].strftime(
                "%Y%m%d"
            )

        if "maxStudyDate" in parameters:
            parameters["maxStudyDate"] = parameters["maxStudyDate"].strftime(
                "%Y%m%d"
            )

        if "patientBirthDate" in parameters:
            parameters["patientBirthDate"] = parameters[
                "patientBirthDate"
            ].strftime("%Y%m%d")

        if "includeFields" in parameters:
            parameters["includeFields"] = ",".join(parameters["includeFields"])

        return parameters


def get_valid_fields(query_level) -> Set[str]:
    """All fields that can be returned at the given MINT query level"""
    if query_level == QueryLevels.INSTANCE:
        return (
            StudyLevel.fields
            | SeriesLevel.fields
            | SeriesLevelPromotable.fields
            | InstanceLevel.fields
        )
    elif query_level == QueryLevels.SERIES:
        return (
            StudyLevel.fields
            | SeriesLevel.fields
            | SeriesLevelPromotable.fields
        )
    elif query_level == QueryLevels.STUDY:
        return StudyLevel.fields
    else:
        raise ValueError(
            f'Unknown query level "{query_level}". Valid values '
            f"are {QueryLevels.ALL}"
        )


class Mint(Searcher):
    """A connection to a mint server"""

    def __init__(self, session, url):
        """
        Parameters
        ----------
        session: requests.session
            A logged in session over which MINT calls can be made
        url: str
            MINT endpoint, including protocol and port. Like https://server:8080/mint
        """

        self.session = session
        self.url = url

    def find_studies(self, query: MintQuery) -> Sequence[MintStudy]:
        """Get all studies matching query

        Parameters
        ----------
        query: MintQuery
            Search based on these parameters. See Query object

        Returns
        -------
        List[MintStudy]
            All studies matching query. Might be empty.
            If Query.QueryLevel is SERIES or INSTANCE, MintStudy objects might
            contain MintSeries and MintInstance instances.
        """
        search_url = self.url + "/studies"
        response = self.session.get(search_url, params=query.as_parameters())
        return parse_mint_studies_response(response.text)


def parse_mint_studies_response(xml_raw) -> List[MintStudy]:
    """Parse the xml response to a MINT find DICOM studies call

    Raises
    ------
    DICOMTrolleyException
        If parsing fails
    """
    try:
        studies = ElementTree.fromstring(xml_raw).findall(
            MintStudy.xml_element
        )
    except ParseError:
        raise DICOMTrolleyException(
            f"Could not parse server response as MINT "
            f"studies. Response was: {xml_raw}"
        )
    return [MintStudy.init_from_element(x) for x in studies]


def parse_attribs(element):
    """Parse xml attributes from a MINT find call to DICOM elements in a Dataset"""
    dataset = Dataset()
    for child in element.findall(MintAttribute.xml_element):
        attr = child.attrib
        val = attr.get("val", "")  # treat missing val as empty
        dataset[attr["tag"]] = DataElement(attr["tag"], attr["vr"], val)

    return dataset


MintInstance.update_forward_refs()  # enables pydantic validation
MintSeries.update_forward_refs()
MintStudy.update_forward_refs()



"""Models the WADO protocol

https://www.dicomstandard.org/dicomweb/retrieve-wado-rs-and-wado-uri/
"""
from concurrent.futures import as_completed
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Sequence

from pydantic.dataclasses import dataclass
from pydicom.dataset import Dataset
from pydicom.errors import InvalidDicomError
from pydicom.filebase import DicomBytesIO
from pydicom.filereader import dcmread
from requests.models import Response
from requests_futures.sessions import FuturesSession

from dicomtrolley.exceptions import DICOMTrolleyException


@dataclass
class InstanceReference:
    """All information needed to download a single slice (SOPInstance) in WADO"""

    study_instance_uid: str
    series_instance_uid: str
    sop_instance_iud: str

    def __str__(self):
        return f"InstanceReference {self.sop_instance_iud}"


class Wado:
    """A connection to a WADO server"""

    def __init__(self, session, url):
        """
        Parameters
        ----------
        session: requests.session
            A logged in session over which WADO calls can be made
        url: str
            WADO endpoint, including protocol and port. Like https://server:8080/wado
        """

        self.session = session
        self.url = url

    @staticmethod
    def to_wado_parameters(instance):
        """WADO url parameters for to retrieve instance

        Returns
        -------
        Dict[str]
            All parameters for a standard WADO get request
        """
        return {
            "requestType": "WADO",
            "studyUID": instance.study_instance_uid,
            "seriesUID": instance.series_instance_uid,
            "objectUID": instance.sop_instance_iud,
            "contentType": "application/dicom",
        }

    @staticmethod
    def parse_wado_response(response: Response) -> Dataset:
        """Create a Dataset out of http response from WADO server

        Raises
        ------
        DICOMTrolleyException
            If response is not as expected or if parsing fails

        Returns
        -------
        Dataset
        """
        if response.status_code != 200:

            raise DICOMTrolleyException(
                f"Calling {response.url} failed ({response.status_code} - "
                f"{response.reason})\n"
                f"response content was {str(response.content[:300])}"
            )
        raw = DicomBytesIO(response.content)
        try:
            return dcmread(raw)
        except InvalidDicomError as e:
            raise DICOMTrolleyException(
                f"Error parsing response as dicom: {e}."
                f" Response content (first 300 elements) was"
                f" {str(response.content[:300])}"
            )

    def get_dataset(self, instance: InstanceReference):
        """Get DICOM dataset for the given instance (slice)

        Raises
        ------
        DICOMTrolleyException
            If getting does not work for some reason

        Returns
        -------
        Dataset
            A pydicom dataset
        """
        return self.parse_wado_response(
            self.session.get(
                self.url, params=self.to_wado_parameters(instance)
            )
        )

    def datasets(self, instances: Sequence[InstanceReference]):
        """Retrieve each instance via WADO

        Returns
        -------
        Iterator[Dataset, None, None]
        """
        for instance in instances:
            yield self.get_dataset(instance)

    def datasets_async(
        self, instances: Sequence[InstanceReference], max_workers=None
    ):
        """Retrieve each instance via WADO

        Parameters
        ----------
        instances: Sequence[InstanceReference]
            Retrieve dataset for each of these instances
        max_workers: int, optional
            Use this number of workers in ThreadPoolExecutor. Defaults to
            default for ThreadPoolExecutor

        Raises
        ------
        DICOMTrolleyException
            When a server response cannot be parsed as DICOM

        Returns
        -------
        Iterator[Dataset, None, None]
        """

        with FuturesSession(
            session=self.session,
            executor=ThreadPoolExecutor(max_workers=max_workers),
        ) as futures_session:
            futures = []
            for instance in instances:
                futures.append(
                    futures_session.get(
                        self.url, params=self.to_wado_parameters(instance)
                    )
                )

            for future in as_completed(futures):
                yield self.parse_wado_response(future.result())



"""Models VNA server specifics. Mainly log-in procedures"""

import requests

from dicomtrolley.dicom_qr import DICOMQR
from dicomtrolley.exceptions import DICOMTrolleyException
from dicomtrolley.mint import Mint
from dicomtrolley.trolley import Trolley
from dicomtrolley.wado import Wado


class VitreaConnection:
    """A server running Vitrea Connection 8.2.0.1"""

    def __init__(self, login_url):
        """

        Parameters
        ----------
        login_url: str
            Full url of login method, including https://
        """
        self.login_url = login_url

    def log_in(self, user, password, realm):
        """
        Parameters
        ----------
        user: str
            username
        password: str
            password to log in with
        realm: str
           realm to log in to

        Returns
        -------
        requests.Session
            A logged in session on the VNA

        Raises
        ------
        DICOMTrolleyException
            If login fails for some reason

        """
        session = requests.Session()
        response = session.post(
            self.login_url,
            headers={
                "X-Userid": user,
                "X-Password": password,
                "X-Realm": realm,
            },
        )
        if response.status_code == 401:
            raise DICOMTrolleyException(
                f"login failed. {response.status_code}:{response.reason}"
            )

        return session

    def get_mint_trolley(
        self, user, password, realm, wado_url, mint_url
    ) -> Trolley:
        """Create a logged in trolley with wado and mint

        Parameters
        ----------
        user: str
            User to log in with
        password:
            Password for user
        realm: str
            realm to log in to
        wado_url: str
            WADO endpoint, including http(s)://  and port
        mint_url
            MINT endpoint, including http(s)://  and port

        Returns
        -------
        Trolley
            logged in trolley with wado and mint

        """
        session = VitreaConnection(login_url=self.login_url).log_in(
            user=user, password=password, realm=realm
        )
        return Trolley(
            wado=Wado(session=session, url=wado_url),
            searcher=Mint(session=session, url=mint_url),
        )


class IMPAXDataCenter:
    """WADO Login for AGFA IMPAX Data Center 3.1.1"""

    def __init__(self, wado_url):
        """

        Parameters
        ----------
        wado_url: str
            Full url of WADO endpoint login method, including https:// and port
        """
        self.wado_url = wado_url

    def log_in(self, user, password):
        """Get a logged in session for WADO on this server

        Parameters
        ----------
        user: str
            username
        password: str
            password to log in with

        Returns
        -------
        requests.Session
            A logged in session on IMPAX

        Raises
        ------
        DICOMTrolleyException
            If login fails for some reason

        """
        session = requests.Session()

        # AGFA IMPAX WADO login is stateful and strange.
        # A better method might be available. However, this one works.
        try:
            session.get(
                self.wado_url
            )  # this is required to make the next GET work..
            response = session.get(
                f"{self.wado_url}/j_security_check"
                f"?j_username={user}&j_password={password}"
            )
        except requests.exceptions.ConnectionError as e:
            raise DICOMTrolleyException(
                f"Error logging in to {self.wado_url}: {e}"
            )

        if (
            "Login Failed!" in response.text
        ):  # login fail status_code is 200 OK..
            raise DICOMTrolleyException(
                f"Logging in to {self.wado_url} failed. "
                f"Invalid Credentials?"
            )
        else:
            return (
                session  # login worked. status_code is now 403 (forbidden)..
            )

    def get_dicom_qr_trolley(
        self,
        wado_user,
        wado_pass,
        host,
        port,
        aet="DICOMTROLLEY",
        aec="ANY-SCP",
    ) -> Trolley:
        """Log in to WADO and create a Trolley with wado and DICOM-QR

        Notes
        -----
        DICOM-QR credentials will only be verified during the first trolley find
        command.

        Parameters
        ----------
        wado_user: str
            User to log in to wado
        wado_pass: str
            Password for wado
        host: str
            Hostname for DICOM-QR
        port: int
            port for DICOM-QR
        aet: str, optional
            Application Entity Title - Name of the calling entity (this class).
            Defaults to 'DICOMTROLLEY'
        aec: str, optional
            Application Entity Called - The name of the server you are calling.
            Defaults to 'ANY-SCP'
        Returns
        -------

        """
        session = IMPAXDataCenter(self.wado_url).log_in(wado_user, wado_pass)
        return Trolley(
            wado=Wado(session, self.wado_url),
            searcher=DICOMQR(host=host, port=port, aet=aet, aec=aec),
        )



"""Provides common base classes that allow different modules to talk to each other."""
from itertools import chain
from typing import Sequence

from pydantic.main import BaseModel
from pydicom.dataset import Dataset

from dicomtrolley.exceptions import DICOMTrolleyException


class DICOMObject(BaseModel):
    """An object in the DICOM world. Base for Study, Series, Instance.

    dicomtrolley search methods always return instances based on DICOMObject
    dicomtrolley download methods take instances based on DICOMObject as input
    """

    uid: str
    data: Dataset

    def __str__(self):
        return type(self).__name__ + " " + self.uid

    def all_instances(self):
        """

        Returns
        -------
        List[Instance]
            All instances contained in this object
        """
        raise NotImplementedError()


class Instance(DICOMObject):
    parent: "Series"

    def all_instances(self):
        """A list containing this instance itself. To match other signatures"""
        return [self]


class Series(DICOMObject):
    instances: Sequence[Instance]
    parent: "Study"

    def all_instances(self):
        """Each instance contained in this series"""
        return self.instances


class Study(DICOMObject):
    series: Sequence[Series]

    def all_instances(self):
        """Return each instance contained in this study"""
        return list(chain(*(x.instances for x in self.series)))


class Searcher:
    """Something that can search for DICOM studies. Base class."""

    def find_studies(self, query) -> Sequence[Study]:
        raise NotImplementedError()

    def find_study(self, query) -> Study:
        """Like find_studies, but guarantees exactly one result. Exception if not.

        This method is meant for searches that contain unique identifiers like
        StudyInstanceUID, AccessionNumber, etc.

        Raises
        ------
        DICOMTrolleyException
            If no results or more than one result is returned by query
        """
        results = self.find_studies(query)
        if len(results) == 0 or len(results) > 1:
            raise DICOMTrolleyException(
                f"Expected exactly one study for query '{query}', but"
                f" found {len(results)}"
            )
        return results[0]


Instance.update_forward_refs()  # enables pydantic validation
Series.update_forward_refs()



"""Fields that can be returned for Study, Series and Instance levels

Notes
-----
Valid fields per level have all been taken from the Vitrea Connection 8.2.0.1
manual. These list might be different for different software.
"""


class InstanceLevel:
    """attributes can be returned if queryLevel=INSTANCE"""

    fields = {
        "SOPInstanceUID",
        "SOPClassUID",
        "Rows",
        "Columns",
        "NumberOfFrames",
        "BitsAllocated",
        "ContentDate",
        "ContentTime",
        "ObservationDateTime",
        "ConceptNameCodeSequence",
        "ContentLabel",
        "ContentDescription",
        "PresentationCreationDate",
        "PresentationCreationTime",
        "ContentCreatorName",
        "RetrieveURI",
        "TransferSyntaxUID",
        "InstanceNumber",
        "Manufacturer",
        "StationName",
        "ManufacturerModelName",
        "DeviceSerialNumber",
        "InstitutionAddress",
        "CorrectedImage",
        "SoftwareVersions",
        "DateOfLastCalibration",
        "TimeOfLastCalibration",
        "PixelPaddingValue",
    }


class SeriesLevel:
    """attributes can be returned if queryLevel=SERIES, or queryLevel=INSTANCE"""

    fields = {
        "SeriesInstanceUID",
        "Modality",
        "SeriesDate",
        "SeriesTime",
        "SeriesDescription",
        "Laterality",
        "AnatomicRegionSequence",
        "BodyPartExamined",
        "FrameOfReferenceUID",
        "PerformedProcedureStepDescription",
        "ProtocolName",
        "PerformingPhysicianName",
        "PerformedProcedureStepStartDate",
        "PerformedProcedureStepStartTime",
        "OperatorsName",
        "PerformedProcedureStepStatus",
        "PresentationIntentType",
        "SeriesNumber",
        "SeriesType",
        "SmallestPixelValueInSeries",
        "SpatialResolution",
        "NumberOfSeriesRelatedInstances",
    }


class SeriesLevelPromotable:
    """Instance level attributes that may be returned even if queryLevel=SERIES

    These attributes are promotable from the Image-level to the Series-level metadata
    if the attribute value for all SOP instances in that series contain the same
    value. Therefore these attributes may be returned in a queryLevel=SERIES query,
    but only when this condition is satisfied.

    Notes
    -----
    Including any of these attributes as return keys in a Series-level query will
    cause the full study metadata to be loaded from disk, even if the corresponding
    attribute was promoted to the study summary MINT metadata. This means there is an
    extra read from disk in addition to the database, so the performance of such
    queries may be slightly slower than queries for other Series-level attributes.
    """

    fields = {
        "Manufacturer",
        "ManufacturerModelName",
        "CorrectedImage",
        "DeviceSerialNumber",
        "InstitutionAddress",
        "SoftwareVersions",
        "StationName",
        "DateOfLastCalibration",
        "TimeOfLastCalibration",
    }


class StudyLevel:
    """All fields that can be returned for queryLevel=STUDY, SERIES or INSTANCE.

    Notes
    -----
    From manual:
    Note that if a specified attribute is not in the study, the attribute will not
    be returned. If the specified attribute is in the study but does not have a
    value, it will be returned without a value.
    """

    fields = {
        "PatientID",
        "IssuerOfPatientID",
        "IssuerOfPatientIDQualifiersSequence",
        "PatientName",
        "PatientSex",
        "PatientBirthDate",
        "PatientBirthTime",
        "StudyInstanceUID",
        "AccessionNumber",
        "IssuerOfAccessionNumberSequence",
        "StudyID",
        "StudyDate",
        "StudyTime",
        "ModalitiesInStudy",
        "ReferringPhysicianName",
        "RequestedProcedureID",
        "StudyDescription",
        "InstanceAvailability",
        "InstitutionName",
        "StudyStatusID",
        "ConfidentialityCode",
        "ProcedureCodeSequence",
        "ReasonForPerformedProcedureCodeSequence",
        "OtherPatientIDsSequence",
        "InstitutionalDepartmentName",
        "NumberOfStudyRelatedSeries",
        "NumberOfStudyRelatedInstances",
        "InstanceAvailability",
        "SOPClassesInStudy",
        "CurrentPatientLocation",
        "SourceApplicationEntityTitle",
    }



"""Combines WADO, MINT and DICOM-QR to make make getting DICOM studies easy.

Notes
-----
Design choices:

WADO, MINT and DICOM-QR modules should be stand-alone. They are not allowed to use
each other's classes. Core has knowledge of all and converts between them if needed

"""
from pathlib import Path
from typing import List, Sequence, Union

from dicomtrolley.core import DICOMObject, Instance, Searcher, Study
from dicomtrolley.wado import InstanceReference, Wado


class Trolley:
    """Combines WADO and MINT or DICOM-QR to make make getting DICOM studies easy."""

    def __init__(self, wado: Wado, searcher: Searcher):
        self.wado = wado
        self.searcher = searcher

    def find_studies(self, query) -> List[Study]:
        """Find study information using MINT.

        Parameters
        ----------
        query:
            Search for these criteria

        Returns
        -------
        List[Study]
        """
        return list(self.searcher.find_studies(query))

    def find_study(self, query) -> Study:
        """Like find study, but returns exactly one result or raises exception.

        For queries using study identifiers such as StudyInstanceUID,AccessionNumber

        Parameters
        ----------
        query
            Search for these criteria. Query documentation for more info.

        Returns
        -------
        Study
        """
        return self.searcher.find_study(query)

    def download(
        self,
        objects: Union[DICOMObject, Sequence[DICOMObject]],
        output_dir,
        use_async=False,
        max_workers=None,
    ):
        """Download the given objects to output dir."""
        if not isinstance(objects, Sequence):
            objects = [objects]
        storage = DICOMStorageDir(output_dir)
        if use_async:
            datasets = self.fetch_all_datasets_async(
                objects=objects, max_workers=max_workers
            )
        else:
            datasets = self.fetch_all_datasets(objects=objects)

        for dataset in datasets:
            storage.save(dataset)

    def fetch_all_datasets(self, objects: Sequence[DICOMObject]):
        """Get full DICOM dataset for each instance in study.

        Returns
        -------
        Iterator[Dataset]
            The downloaded dataset and the context that was used to download it
        """
        yield from self.wado.datasets(self.extract_instances(objects))

    @staticmethod
    def extract_instances(
        objects: Sequence[Union[DICOMObject, InstanceReference]]
    ):
        """Get all individual instances from input.

        A pre-processing step for getting datasets.

        Parameters
        ----------
        objects: list[Union[DICOMObject, InstanceReference]]
            Any combination of Study, Series and Instance objects

        Returns
        -------
        List[InstanceReference]
            A reference to each instance (slice)
        """
        instances = []
        for item in objects:
            if isinstance(item, InstanceReference):
                instances.append(item)
            else:
                instances = instances + [
                    to_wado_reference(x) for x in item.all_instances()
                ]
        return instances

    def fetch_all_datasets_async(self, objects, max_workers=None):
        """Get DICOM dataset for each instance given objects using multiple threads.

        Parameters
        ----------
        objects: List[DICOMObject]
            get dataset for each instance contained in these objects
        max_workers: int, optional
            Max number of ThreadPoolExecutor workers to use. Defaults to
            ThreadPoolExecutor default

        Raises
        ------
        DICOMTrolleyException
            If getting or parsing of any instance fails

        Returns
        -------
        Iterator[Dataset, None, None]
            The downloaded dataset and the context that was used to download it
        """

        yield from self.wado.datasets_async(
            instances=self.extract_instances(objects),
            max_workers=max_workers,
        )


def to_wado_reference(instance: Instance) -> InstanceReference:
    """Simplify a more extensive MINT instance to an InstanceReference.

    needed for calls to WADO functions
    """
    return InstanceReference(
        study_instance_uid=instance.parent.parent.uid,
        series_instance_uid=instance.parent.uid,
        sop_instance_iud=instance.uid,
    )


class DICOMStorageDir:
    """A directory that you can write datasets to."""

    def __init__(self, path: str):
        self.path = path

    def __str__(self):
        return f"DICOMStorageDir at {self.path}"

    def save(self, dataset):
        """Write dataset. Creates sub-folders if needed."""

        slice_path = Path(self.path) / self.generate_path(dataset)
        slice_path.parent.mkdir(parents=True, exist_ok=True)
        dataset.save_as(slice_path)

    def generate_path(self, dataset):
        """A path studyid/seriesid/instanceid to save a slice to."""

        stu_uid = self.get_value(dataset, "StudyInstanceUID")
        ser_uid = self.get_value(dataset, "SeriesInstanceUID")
        sop_uid = self.get_value(dataset, "SOPInstanceUID")
        return Path(self.path) / stu_uid / ser_uid / sop_uid

    @staticmethod
    def get_value(dataset, tag_name):
        """Extract value for use in path. If not found return default."""
        default = "unknown"
        return str(dataset.get(tag_name, default)).replace(".", "_")



"""Models parsing things into study/series/instance structure"""

from typing import Any, DefaultDict, List

from pydicom.dataset import Dataset

from dicomtrolley.core import Instance, Series, Study
from dicomtrolley.exceptions import DICOMTrolleyException


class TreeNode(DefaultDict[Any, "TreeNode"]):
    """Recursive defaultdict with a 'data' property. Helps parse to tree structures.

    Examples
    --------
    >>> root = TreeNode()
    >>> root['study1']['series1']['instance1'].data = 'some instance info'
    >>> 'study1' in root
    True
    >>> root['study1']['series2'].data = 'some series data'
    >>> list(root['study1'].keys)
    ['series1', 'series2']
    """

    def __init__(self, data=None, allow_overwrite=True):
        """

        Parameters
        ----------
        data:
            Optional data to associate with this node
        allow_overwrite: bool, optional
            If False, will raise exception when overwriting data attribute
        """
        super().__init__(lambda: TreeNode(allow_overwrite=allow_overwrite))
        self._data = data
        self.allow_overwrite = allow_overwrite

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        if self.allow_overwrite or not self.data:
            self._data = value
        else:
            raise ValueError("Overwriting data is not allowed")


class DICOMParseTree:
    """Models study/series/instance as a tree. Allows arbitrary branch insertions"""

    def __init__(self, root=None, allow_overwrite=False):
        """

        Parameters
        ----------
        root: TreeNode, optional
            Root node. Will contain all. Defaults to empty
        allow_overwrite: bool, optional
            If False, will raise exception when overwriting any TreeNode.data,
            Defaults to False
        """
        if not root:
            self.root = TreeNode(allow_overwrite=allow_overwrite)

    def insert(self, data, study, series=None, instance=None):
        """Insert data at the given level in tree. Will complain about missing
        branches

        Raises
        ------
        DICOMTrolleyException
            If inserting fails for any reason
        """
        if instance and not series:
            DICOMTrolleyException(
                f"Instance was given ({instance}) but series was not. I can "
                f"not insert this into a study/series/instance tree"
            )
        try:
            if series:
                if instance:
                    self.root[study][series][instance].data = data
                else:
                    self.root[study][series].data = data
            else:
                self.root[study].data = data
        except ValueError as e:
            raise DICOMTrolleyException(
                f"Error inserting dataset into {study}/{series}/{instance}: {e}"
            )

    def insert_dataset(self, ds: Dataset):
        self.insert(
            data=ds,
            study=ds.get("StudyInstanceUID"),
            series=ds.get("SeriesInstanceUID"),
            instance=ds.get("SOPInstanceUID"),
        )

    def as_studies(self) -> List[Study]:
        """Tree as DICOMObject instances: Study containing Series, Instance"""

        def value_or_dataset(val):
            if val:
                return val
            else:
                return Dataset()

        studies = []
        for study_instance_uid, study_node in self.root.items():
            study = Study(
                uid=study_instance_uid,
                data=value_or_dataset(study_node.data),
                series=tuple(),
            )
            studies.append(study)
            all_series = []
            for series_instance_uid, series_node in study_node.items():
                series = Series(
                    uid=series_instance_uid,
                    data=value_or_dataset(series_node.data),
                    parent=study,
                    instances=tuple(),
                )
                all_series.append(series)
                all_instances = []
                for sop_instance_uid, instance_node in series_node.items():
                    all_instances.append(
                        Instance(
                            uid=sop_instance_uid,
                            data=value_or_dataset(instance_node.data),
                            parent=series,
                        )
                    )
                series.instances = tuple(all_instances)
            study.series = tuple(all_series)
        return studies
